import { ValidationContext } from "../ValidationContext";
import { GltfData } from "./GltfData";
/**
 * A class for validating the `EXT_instance_features` extension in
 * glTF assets.
 *
 * This class assumes that the structure of the glTF asset itself
 * has already been validated (e.g. with the glTF Validator).
 *
 * @internal
 */
export declare class ExtInstanceFeaturesValidator {
    /**
     * Performs the validation to ensure that the `EXT_instance_features`
     * extensions in the given glTF are valid
     *
     * @param path - The path for validation issues
     * @param gltfData - The glTF data, containing the parsed JSON and the
     * (optional) binary buffer
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    static validateGltf(path: string, gltfData: GltfData, context: ValidationContext): Promise<boolean>;
    /**
     * Validate the given EXT_instance_features extension object that was
     * found in the given node
     *
     * This assumes that the given object has already been validated
     * to the extent that is checked by the `FeatureIdValidator`,
     * with the `validateCommonFeatureId` method.
     *
     * @param path - The path for validation issues
     * @param instanceFeatures - The EXT_instance_features extension object
     * @param meshGpuInstancing - The EXT_mesh_gpu_instancing extension object
     * @param gltfData - The glTF data
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    private static validateExtInstanceFeatures;
    /**
     * Validate the given feature ID object that was found in the
     * `featureIds` array of an EXT_instance_features extension object
     *
     * @param path - The path for validation issues
     * @param featureId - The feature ID
     * @param meshGpuInstancing - The `EXT_mesh_gpu_instancing` extension object
     * that contains the attribute definitions
     * @param gltfData - The glTF data
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    private static validateInstanceFeaturesFeatureId;
    /**
     * Validate the given feature ID `attribute` value that was found in
     * a feature ID definition
     *
     * @param path - The path for validation issues
     * @param attribute - The attribute (i.e. the supposed number that
     * will be used for the `_FEATURE_ID_${attribute}` attribute name)
     * @param featureCount - The `featureCount` value from the feature ID definition
     * @param meshGpuInstancing - The `EXT_mesh_gpu_instancing` extension object
     * that contains the attribute definitions
     * @param gltfData - The glTF data
     * @param propertyTableState - The validation state of the property table
     * definition (i.e. the index into the property tables array)
     * @param nullFeatureId - The `nullFeatureId` of the `featureId` object
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    private static validateFeatureIdAttribute;
}
//# sourceMappingURL=ExtInstanceFeaturesValidator.d.ts.map