import { ValidationContext } from "../ValidationContext";
import { GltfData } from "./GltfData";
/**
 * A class for validating the `EXT_mesh_features` extension in
 * glTF assets.
 *
 * This class assumes that the structure of the glTF asset itself
 * has already been validated (e.g. with the glTF Validator).
 *
 * @internal
 */
export declare class ExtMeshFeaturesValidator {
    /**
     * Performs the validation to ensure that the `EXT_mesh_features`
     * extensions in the given glTF are valid
     *
     * @param path - The path for validation issues
     * @param gltfData - The glTF data, containing the parsed JSON and the
     * (optional) binary buffer
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    static validateGltf(path: string, gltfData: GltfData, context: ValidationContext): Promise<boolean>;
    /**
     * Validate the given EXT_mesh_features extension object that was
     * found in the given mesh primitive.
     *
     * @param path - The path for validation issues
     * @param meshFeatures - The EXT_mesh_features extension object
     * @param meshPrimitive - The mesh primitive that contains the extension
     * @param gltfData - The glTF data
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    private static validateExtMeshFeatures;
    /**
     * Validate the given feature ID object that was found in the
     * `featureIds` array of an EXT_mesh_features extension object.
     *
     * This assumes that the given object has already been validated
     * to the extent that is checked by the `FeatureIdValidator`,
     * with the `validateCommonFeatureId` method.
     *
     * @param path - The path for validation issues
     * @param featureId - The feature ID
     * @param meshPrimitive - The mesh primitive that contains the extension
     * @param gltfData - The glTF data
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    private static validateMeshFeaturesFeatureId;
    /**
     * Validate the given feature ID `attribute` value that was found in
     * a feature ID definition
     *
     * @param path - The path for validation issues
     * @param attribute - The attribute (i.e. the supposed number that
     * will be used for the `_FEATURE_ID_${attribute}` attribute name)
     * @param featureCount - The `featureCount` value from the feature ID definition
     * @param meshPrimitive - The mesh primitive that contains the extension
     * @param gltfData - The glTF data
     * @param propertyTableState - The validation state of the property table
     * definition (i.e. the index into the property tables array)
     * @param nullFeatureId - The `nullFeatureId` of the `featureId` object
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    private static validateFeatureIdAttribute;
    /**
     * Validate the given feature ID `texture` value that was found in
     * a feature ID definition
     *
     * @param path - The path for validation issues
     * @param featureIdTexture - The feature ID texture definition
     * @param featureCount - The `featureCount` value from the feature ID definition
     * @param meshPrimitive - The mesh primitive that contains the extension
     * @param gltfData - The glTF data
     * @param propertyTableState - The validation state of the property table
     * definition (i.e. the index into the property tables array)
     * @param nullFeatureId - The `nullFeatureId` of the `featureId` object
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    private static validateFeatureIdTexture;
    /**
     * Validate the data of the given feature ID texture.
     *
     * This will try to read the image data, check whether it matches
     * the `channels` definition of the feature ID texture, and whether
     * the number of feature IDs (created from the respective channels
     * of the image pixels) actually matches the given `featureCount`.
     *
     * @param path - The path for validation issues
     * @param featureIdTexture - The feature ID texture
     * @param featureCount - The `featureCount` value from the feature ID definition
     * @param gltfData - The glTF data
     * @param propertyTableState - The validation state of the property table
     * definition (i.e. the index into the property tables array)
     * @param nullFeatureId - The `nullFeatureId` of the `featureId` object
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    private static validateFeatureIdTextureData;
}
//# sourceMappingURL=ExtMeshFeaturesValidator.d.ts.map