import { ValidationContext } from "../ValidationContext";
import { ValidatedElement } from "../ValidatedElement";
import { GltfData } from "./GltfData";
/**
 * A class for validation functionality related to feature IDs, as they
 * appear in the `EXT_mesh_features` and `EXT_instance_features`
 * extensions.
 *
 * @internal
 */
export declare class FeatureIdValidator {
    /**
     * Validate the common elements of a feature ID object.
     *
     * This refers to `featureId` objects that are found in the
     * `EXT_mesh_features` and `EXT_instance_features` extension
     * objects.
     *
     * It ensures that...
     * - the value being an object
     * - the nullFeatureId (if present) being valid
     * - the label (if present) being valid
     * - the featureCount being present and valid
     *
     * It does NOT validate the `texture` or `attribute` properties
     * that may be found in the object, depending on whether it is
     * part of the `EXT_mesh_features` or `EXT_instance_features`
     * extension object.
     *
     * @param path - The path for validation issues
     * @param featureId - The feature ID
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    static validateCommonFeatureId(path: string, featureId: any, context: ValidationContext): boolean;
    /**
     * Validate the given set of feature ID values that have either
     * been found in an feature ID texture or in a feature ID attribute.
     *
     * This will check the validity of the 'featureCount' for the
     * given set of features, depending on the presence of the
     * 'nullFeatureId', and whether the feature IDs are valid
     * indices into a property table (if the property table count
     * was given)
     *
     * @param path - The path for validation issues
     * @param sourceName - The source, 'texture' or 'attribute'
     * @param featureIdSet - The feature ID set. Note that This set
     * might be modifified by this method!
     * @param featureCount - The `featureCount` value from the feature ID definition
     * @param propertyTableState - The validation state of the property table
     * definition (i.e. the index into the property tables array)
     * @param nullFeatureId - The `nullFeatureId` of the `featureId` object
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    static validateFeatureIdSet(path: string, sourceName: string, featureIdSet: Set<number>, featureCount: number, propertyTableState: ValidatedElement<{
        count: number;
    }>, nullFeatureId: number | undefined, context: ValidationContext): boolean;
    /**
     * Obtain the `count` of the property table that is referred to
     * with the given index.
     *
     * This assumes that the validity of this index has already been
     * checked with `validateFeatureIdPropertyTable`. If any
     * element that leads to the `count` is invalid or not defined,
     * then `undefined` will be returned.
     *
     * @param propertyTableIndex - The value that was found as the `propertyTable`
     * in the definition, indicating the index into the property tables array
     * @param gltfData - The glTF data
     * @returns The `count` of the property table
     */
    static obtainPropertyTableCount(propertyTableIndex: number, gltfData: GltfData): number | undefined;
}
//# sourceMappingURL=FeatureIdValidator.d.ts.map