/// <reference types="node" />
import { ValidationContext } from "../ValidationContext";
import { GltfData } from "./GltfData";
/**
 * A class for reading `GltfData` from a buffer.
 *
 * This is only used internally, to resolve all elements that are
 * part of a `GltfData` that is required for the validation.
 *
 * @internal
 */
export declare class GltfDataReader {
    /**
     * Read the glTF data from the given buffer.
     *
     * This currently supports binary glTF and embedded glTF assets.
     *
     * If the data can not be read, then the appropriate issue
     * will be added to the given context, and `undefined` will
     * be returned
     *
     * @param path - The path for `ValidationIssue` instances
     * @param input - The raw glTF data
     * @param context - The `ValidationContext`
     * @returns The glTF data, or `undefined`
     */
    static readGltfData(path: string, input: Buffer, context: ValidationContext): Promise<GltfData | undefined>;
    /**
     * Read the binary glTF data from the given buffer.
     *
     * If the data can not be read, then the appropriate issue
     * will be added to the given context, and `undefined` will
     * be returned
     *
     * @param path - The path for `ValidationIssue` instances
     * @param input - The raw glTF data
     * @param context - The `ValidationContext`
     * @returns The glTF data, or `undefined`
     */
    private static readBinaryGltfData;
    /**
     * Return the binary buffer data that is described by
     * the buffers/bufferViews of the given glTF, referring
     * to the given binary buffer data.
     *
     * @param gltf - The glTF object
     * @param binary - The binary buffer data
     * @param resourceResolver - The resource resolver
     * @returns The binary buffer data
     */
    private static resolveBinaryBufferData;
    /**
     * Try to read a glTF-Tranform document from the given GLB buffer
     * data, returning `undefined` if the document cannot be read.
     *
     * @param input - The input GLB buffer
     * @returns The document, or `undefined`
     */
    private static readBinaryGltfDocument;
    /**
     * Read the glTF data from the given buffer, which contains
     * glTF JSON.
     *
     * This currently only supports embedded glTF.
     *
     * If the data can not be read, then the appropriate issue
     * will be added to the given context, and `undefined` will
     * be returned
     *
     * @param path - The path for `ValidationIssue` instances
     * @param input - The raw glTF data
     * @param context - The `ValidationContext`
     * @returns The glTF data, or `undefined`
     */
    private static readJsonGltfData;
    /**
     * Try to read a glTF-Tranform document from the given glTF JSON
     * data, returning `undefined` if the document cannot be read.
     *
     * @param input - The input buffer containging glTF JSON
     * @returns The document, or `undefined`
     */
    private static readJsonGltfDocument;
}
//# sourceMappingURL=GltfDataReader.d.ts.map