/// <reference types="node" />
import { BinaryBufferData } from "3d-tiles-tools";
import { ImageData } from "./ImageData";
import { ValidationContext } from "../ValidationContext";
/**
 * A class for reading and accessing `ImageData` instances
 *
 * @internal
 */
export declare class ImageDataReader {
    /**
     * Read the `ImageData` from the given glTF image.
     *
     * This will read the image data from the `uri` of the given image,
     * or from the `bufferView` of the image.
     *
     * If the data cannot be read, then a validation error will be
     * added to the given context, and `undefined` will be returned.
     *
     * @param path - The path for the `ValidationIssue` instances
     * @param image - The glTF `image` object
     * @param binaryBufferData - The `BinaryBufferData` that contains
     * the buffer view data of the glTF
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns The image data, or `undefined`
     */
    static readFromImage(path: string, image: any, binaryBufferData: BinaryBufferData, context: ValidationContext): Promise<ImageData | undefined>;
    /**
     * Obtains the value from the specified pixel and channels in the
     * given image data.
     *
     * This is computed by composing the (supposedly `UINT8`) channel
     * values of the pixels
     *
     * @param imageData - The ImageData object
     * @param x - The x-coordinate
     * @param y - The y-coordinate
     * @param channels - The `channels` definition from the texture
     * @returns The value
     */
    static getValue(imageData: ImageData, x: number, y: number, channels: number[]): number;
    /**
     * Returns the value of the specified channel of the specified pixel
     * in the given image data.
     *
     * @param imageData - The ImageData object
     * @param x - The x-coordinate
     * @param y - The y-coordinate
     * @param channel - The `channel` (usually in [0...3] for RGBA)
     * @returns The value
     */
    static getChannelValue(imageData: ImageData, x: number, y: number, channel: number): number;
    /**
     * Try to read image data from the given buffer.
     *
     * The exact set of image formats that may be contained in the given
     * buffer is not specified. But it will support PNG and JPEG.
     *
     * @param imageDataBuffer - The image data buffer
     * @returns The `ImageData`
     * @throws An error if the image data can not be read
     */
    static readUnchecked(imageDataBuffer: Buffer): Promise<ImageData>;
}
//# sourceMappingURL=ImageDataReader.d.ts.map