"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PropertyAttributeValidator = void 0;
const _3d_tiles_tools_1 = require("3d-tiles-tools");
const _3d_tiles_tools_2 = require("3d-tiles-tools");
const BasicValidator_1 = require("../BasicValidator");
const RootPropertyValidator_1 = require("../RootPropertyValidator");
const ExtendedObjectsValidators_1 = require("../ExtendedObjectsValidators");
const MetadataStructureValidator_1 = require("../metadata/MetadataStructureValidator");
const PropertyAttributePropertyValidator_1 = require("./PropertyAttributePropertyValidator");
/**
 * A class for validations related to `propertyAttribute` objects.
 *
 * This class performs the basic JSON-level validation of the
 * property attributes.
 *
 * The validation of any of the underlying binary data of
 * a property attribute has to start at the mesh primitive
 * that refers to the property attribute, because it requires
 * knowledge about the attributes that are defined in the
 * referring mesh primitive.
 *
 * @internal
 */
class PropertyAttributeValidator {
    /**
     * Performs the validation to ensure that the given object is a
     * valid `propertyAttribute` object.
     *
     * @param path - The path for the `ValidationIssue` instances
     * @param propertyAttribute - The object to validate
     * @param schema - The `Schema` object
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    static validatePropertyAttribute(path, propertyAttribute, schema, context) {
        // Make sure that the given value is an object
        if (!BasicValidator_1.BasicValidator.validateObject(path, "propertyAttribute", propertyAttribute, context)) {
            return false;
        }
        let result = true;
        // Validate the object as a RootProperty
        if (!RootPropertyValidator_1.RootPropertyValidator.validateRootProperty(path, "propertyAttribute", propertyAttribute, context)) {
            result = false;
        }
        // Perform the validation of the object in view of the
        // extensions that it may contain
        if (!ExtendedObjectsValidators_1.ExtendedObjectsValidators.validateExtendedObject(path, propertyAttribute, context)) {
            result = false;
        }
        // If there was an extension validator that overrides the
        // default validation, then skip the remaining validation.
        if (ExtendedObjectsValidators_1.ExtendedObjectsValidators.hasOverride(propertyAttribute)) {
            return result;
        }
        // Validate that the class and properties are structurally
        // valid and comply to the metadata schema
        const className = propertyAttribute.class;
        const attributeProperties = propertyAttribute.properties;
        if (!MetadataStructureValidator_1.MetadataStructureValidator.validateMetadataStructure(path, "property attribute", className, attributeProperties, schema, context)) {
            // Bail out early if the structure is not valid!
            return false;
        }
        // Validate the name.
        // If the name is defined, it MUST be a string.
        if (!BasicValidator_1.BasicValidator.validateOptionalString(path, propertyAttribute, "name", context)) {
            result = false;
        }
        // Here, the basic structure of the class and properties
        // have been determined to be valid. Continue to validate
        // the values of the properties.
        const validProperties = (0, _3d_tiles_tools_2.defaultValue)(attributeProperties, {});
        const validPropertyNames = Object.keys(validProperties);
        const classes = (0, _3d_tiles_tools_2.defaultValue)(schema.classes, {});
        const metadataClass = classes[className];
        const classProperties = (0, _3d_tiles_tools_2.defaultValue)(metadataClass.properties, {});
        // Validate each property
        for (const propertyName of validPropertyNames) {
            const propertyPath = path + "/properties/" + propertyName;
            const classProperty = classProperties[propertyName];
            let enumValueType = undefined;
            const enumType = classProperty.enumType;
            if ((0, _3d_tiles_tools_1.defined)(enumType)) {
                const enums = (0, _3d_tiles_tools_2.defaultValue)(schema.enums, {});
                const metadataEnum = enums[enumType];
                enumValueType = metadataEnum.valueType || "UINT16";
            }
            // Note: The check whether 'required' properties are
            // present and have values was already done by the
            // MetadataStructureValidator
            const propertyValue = validProperties[propertyName];
            if ((0, _3d_tiles_tools_1.defined)(propertyValue)) {
                if (!PropertyAttributePropertyValidator_1.PropertyAttributePropertyValidator.validatePropertyAttributeProperty(propertyPath, propertyName, propertyValue, classProperty, enumValueType, context)) {
                    result = false;
                }
            }
        }
        return result;
    }
}
exports.PropertyAttributeValidator = PropertyAttributeValidator;
