import { Schema } from "3d-tiles-tools";
import { ValidationContext } from "../ValidationContext";
import { GltfData } from "./GltfData";
/**
 * A class for the validation of values that are stored
 * in property attributes.
 *
 * The methods in this class assume that the structural
 * validity of the input objects has already been checked
 * by a `PropertyAttributeValidator`.
 *
 * @internal
 */
export declare class PropertyAttributeValuesValidator {
    /**
     * Performs the validation to ensure that the specified property
     * attribute contains valid values.
     *
     * This is supposed to be called after the validity of the top-level
     * extension object, the schema, and the property attribute itself have
     * been checked (the latter with
     * `PropertyAttributeValidator.validatePropertyAttribute`).
     *
     * It assumes that they are structurally valid, and ONLY checks the
     * validity of the values in the context of the mesh primitive
     * that refers to the property attribute.
     *
     * @param path - The path for the `ValidationIssue` instances
     * @param propertyAttributeIndex - The index that was found as
     * an element `propertyAttributes[i]` of the extension object
     * that was found in the extension object in the mesh primitive
     * @param meshPrimitive - The glTF mesh primitive that contained
     * the extension object
     * @param meshIndex - The index of the mesh (only for details
     * in validation messages)
     * @param pimitiveIndex - The index of the primitive (only for details
     * in validation messages)
     * @param schema - The metadata schema
     * @param gltfStructuralMetadata - The top-level glTF structural
     * metadata object
     * @param gltfData - The glTF data
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the values in the object have been valid
     */
    static validatePropertyAttributeValues(path: string, propertyAttributeIndex: number, meshPrimitive: any, meshIndex: number, primitiveIndex: number, schema: Schema, gltfStructuralMetadata: any, gltfData: GltfData, context: ValidationContext): Promise<boolean>;
    /**
     * Validate the values of a single property of a property attribute.
     *
     * This assumes that the gltfData contains a valid gltfDocument.
     *
     * @param path - The path for `ValidationIssue` instances
     * @param propertyName - The property name
     * @param propertyAttributeProperty - The property attribute property
     * @param meshPrimitive - The mesh primitive
     * @param meshIndex - The index of the mesh (only for details
     * in validation messages)
     * @param pimitiveIndex - The index of the primitive (only for details
     * in validation messages)
     * @param schema  - The metadata schema
     * @param metadataClassName - Te class name that was given in the
     * surrounding property attribute
     * @param gltfData - The glTF data
     * @param context - The `ValidationContext`
     * @returns Whether the property is valid
     */
    private static validatePropertyAttributePropertyValues;
}
//# sourceMappingURL=PropertyAttributeValuesValidator.d.ts.map