"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PropertyTableDefinitionValidator = void 0;
const _3d_tiles_tools_1 = require("3d-tiles-tools");
const GltfExtensionValidationIssues_1 = require("../../issues/GltfExtensionValidationIssues");
const BasicValidator_1 = require("../BasicValidator");
/**
 * A class for the validation of a single property table that
 * is referred to by a feature ID definition.
 *
 */
class PropertyTableDefinitionValidator {
    /**
     * Validate the given feature ID `propertyTable` value that was found in
     * a feature ID definition.
     *
     * The returned object will contain two properties:
     * - `wasPresent`: Whether a propertyTable (index) was given
     * - `validatedElement`: The validated property table object, only
     *    insofar that it contains a defined `count` value
     *
     * This will check whether the `propertyTable` refers to an exising
     * property table in the `EXT_structural_metadata` extension object,
     * and this property table has a valid `count`.
     *
     * It will NOT check the validity of the property table itself. This
     * will be done by the `EXT_structural_metadata` validator.
     *
     * @param path - The path for validation issues
     * @param propertyTableIndex - The value that was found as the `propertyTable`
     * in the definition, indicating the index into the property tables array
     * @param gltfData - The glTF data
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the state summarizing the definition
     */
    static validatePropertyTableDefinition(path, propertyTableIndex, gltfData, context) {
        // Return immediately when there are no property table
        const propertyTableState = {
            wasPresent: false,
            validatedElement: undefined,
        };
        if (!(0, _3d_tiles_tools_1.defined)(propertyTableIndex)) {
            return propertyTableState;
        }
        propertyTableState.wasPresent = true;
        const gltf = gltfData.gltf;
        const extensions = gltf.extensions || {};
        const structuralMetadata = extensions["EXT_structural_metadata"];
        if (!structuralMetadata) {
            const message = `The feature ID refers to a property table with index ` +
                `${propertyTableIndex}, but the glTF did not contain an ` +
                `'EXT_structural_metadata' extension object`;
            const issue = GltfExtensionValidationIssues_1.GltfExtensionValidationIssues.INVALID_GLTF_STRUCTURE(path, message);
            context.addIssue(issue);
            return propertyTableState;
        }
        const propertyTables = structuralMetadata.propertyTables;
        if (!propertyTables) {
            const message = `The feature ID refers to a property table with index ` +
                `${propertyTableIndex}, but the 'EXT_structural_metadata' ` +
                `extension object did not define property tables`;
            const issue = GltfExtensionValidationIssues_1.GltfExtensionValidationIssues.INVALID_GLTF_STRUCTURE(path, message);
            context.addIssue(issue);
            return propertyTableState;
        }
        // Validate the propertyTable(Index)
        // The propertyTable MUST be an integer in [0,numPropertyTables)
        const numPropertyTables = (0, _3d_tiles_tools_1.defaultValue)(propertyTables.length, 0);
        if (!BasicValidator_1.BasicValidator.validateIntegerRange(path, "propertyTable", propertyTableIndex, 0, true, numPropertyTables, false, context)) {
            return propertyTableState;
        }
        const propertyTable = propertyTables[propertyTableIndex];
        if (!propertyTable) {
            // An issue will be added to the validation context by
            // the `EXT_structural_metadata` validation
            return propertyTableState;
        }
        const count = propertyTable.count;
        if (count === undefined) {
            // An issue will be added to the validation context by
            // the `EXT_structural_metadata` validation
            return propertyTableState;
        }
        propertyTableState.validatedElement = propertyTable;
        return propertyTableState;
    }
}
exports.PropertyTableDefinitionValidator = PropertyTableDefinitionValidator;
