"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PropertyTextureEnumMetadataPropertyModel = void 0;
const _3d_tiles_tools_1 = require("3d-tiles-tools");
const _3d_tiles_tools_2 = require("3d-tiles-tools");
const ImageDataReader_1 = require("./ImageDataReader");
/**
 * Implementation of a MetadataPropertyModel for a single
 * property texture property that contains enum values.
 *
 * @internal
 */
class PropertyTextureEnumMetadataPropertyModel {
    /**
     * Creates a new instance
     *
     * @param imageData - The image data
     * @param propertyTextureProperty - The property texture property
     * @param classProperty - The class property
     * @param enumValueType - The `valueType` of the enum type of
     * the given class property
     * @param valueValueNames - The mapping from enum value values
     * to enum value names for the enum type of the given class
     * property
     */
    constructor(imageData, propertyTextureProperty, classProperty, enumValueType, enumValueValueNames) {
        this.imageData = imageData;
        this.propertyTextureProperty = propertyTextureProperty;
        this.classProperty = classProperty;
        this.enumValueType = enumValueType;
        this.enumValueValueNames = enumValueValueNames;
    }
    /** {@inheritDoc MetadataPropertyModel.getPropertyValue} */
    getPropertyValue(pixelCoordinates) {
        const classProperty = this.classProperty;
        const value = this.getRawPropertyValue(pixelCoordinates);
        const valueNames = this.enumValueValueNames;
        const processedValue = _3d_tiles_tools_1.MetadataValues.processNumericEnumValue(classProperty, valueNames, value);
        return processedValue;
    }
    /** {@inheritDoc MetadataPropertyModel.getRawPropertyValue} */
    getRawPropertyValue(pixelCoordinates) {
        const propertyTextureProperty = this.propertyTextureProperty;
        const classProperty = this.classProperty;
        const channels = propertyTextureProperty.channels;
        // First, collect the actual values from the image
        // data in a buffer
        const x = pixelCoordinates[0];
        const y = pixelCoordinates[1];
        const buffer = Buffer.alloc(channels.length);
        for (let c = 0; c < channels.length; c++) {
            const channelValue = ImageDataReader_1.ImageDataReader.getChannelValue(this.imageData, x, y, c);
            buffer.writeUInt8(channelValue & 0xff, c);
        }
        // The enum value type is assumed to be defined
        // for enum properties
        const enumValueType = this.enumValueType;
        // Handle enum array properties
        if (classProperty.array) {
            // Variable-length arrays are not supported for property
            // textures, so the `count` must be defined here
            const count = classProperty.count;
            const value = Array(count);
            for (let e = 0; e < count; e++) {
                const element = _3d_tiles_tools_2.NumericBuffers.getNumericFromBuffer(buffer, e, enumValueType);
                value[e] = element;
            }
            return value;
        }
        // Handle enum (non-array) properties
        const value = _3d_tiles_tools_2.NumericBuffers.getNumericFromBuffer(buffer, 0, enumValueType);
        return Number(value);
    }
}
exports.PropertyTextureEnumMetadataPropertyModel = PropertyTextureEnumMetadataPropertyModel;
