import { ClassProperty } from "3d-tiles-tools";
import { ValidationContext } from "../ValidationContext";
/**
 * A class for validations related to `propertyTexture.property` objects.
 *
 * @internal
 */
export declare class PropertyTexturePropertyValidator {
    /**
     * Performs the validation to ensure that the given object is a
     * valid `propertyTexture.property` object.
     *
     * This checks the basic validity of the property, namely
     * - it is an object
     * - its `index` refers to an existing texture in the glTF
     * - its `texCoord` is a nonnegative integer
     * - its `chennels` (if defined) is an array of at least one nonnegative integers
     * - its type (defined via the `classProperty`) is suitable for a property texture
     * - the sampler of the texture is suitable for a property texture
     * - its `offset/scale/min/max` properties match the `classProperty` structure
     *
     * Checking the validity of the `texCoord` for a mesh primitive, and
     * checking the values (whether they are valid enum values, or whether
     * their min/max are valid) is done by the PropertyTextureValuesValidator.
     *
     * @param path - The path for the `ValidationIssue` instances
     * @param propertyName - The name of the property
     * @param propertyTextureProperty - The object to validate
     * @param gltf - The containing glTF object
     * @param meshPrimitive - The mesh primitive that contains the extension
     * @param classProperty - The `ClassProperty` definition from the schema
     * @param enumValueType - The value type of the enum that is represented
     * with the given class property, or `undefined` if it is not an enum
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    static validatePropertyTextureProperty(path: string, propertyName: string, propertyTextureProperty: any, gltf: any, classProperty: ClassProperty, enumValueType: string | undefined, context: ValidationContext): boolean;
    /**
     * Validates that the given number of channels matches the
     * number of bytes of the type of the given `classProperty`.
     *
     * The given number is the length of the `channels` array of
     * a property texture property (if it is defined). If the
     * `channels` array is not defined, then this value should be
     * `1`, because the default value for the `channels` is `[0]`.
     *
     * This assumes that the given `classProperty` does not have
     * the type `STRING`, and has already been determined to be
     * structurally valid.
     *
     * If the number is not valid, then a validation error will be added
     * to the given context, and `false` will be returned.
     *
     * @param path - The path for the `ValidationIssue` instances
     * @param propertyName - The name of the property
     * @param numberOfChannels - The number of channels
     * @param classProperty - The `ClassProperty` definition from the schema
     * @param enumValueType - The value type of the enum that is represented
     * with the given class property, or `undefined` if it is not an enum
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the number was valid
     */
    static validatePropertyTexturePropertyTypeSize(path: string, propertyName: string, numberOfChannels: number, classProperty: ClassProperty, enumValueType: string | undefined, context: ValidationContext): boolean;
    /**
     * Validates that the given `classProperty` is basically suitable
     * for a property texture (meaning that it is not a variable-length
     * array and not of type STRING).
     *
     * If the type is not valid, then a validation error will be added
     * to the given context, and `false` will be returned.
     *
     * @param path - The path for the `ValidationIssue` instances
     * @param propertyName - The name of the property
     * @param classProperty - The `ClassProperty` definition from the schema
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    static validateClassPropertyForPropertyTextureProperty(path: string, propertyName: string, classProperty: ClassProperty, context: ValidationContext): boolean;
}
//# sourceMappingURL=PropertyTexturePropertyValidator.d.ts.map