import { ValidationContext } from "../ValidationContext";
/**
 * A class for validating feature ID or property textures in the
 * glTF metadata extensions.
 *
 * @internal
 */
export declare class TextureValidator {
    /**
     * Validate the `texCoord` definition of a texture.
     *
     * Without further context, this can only check whether the `texCoord`
     * property is a nonnegative integer.
     *
     * It can NOT check whether the `texCoord` results in a
     * `TEXCOORD_${texCoord}` attribute that actually appears in a
     * mesh primitive. For this, `validateTexCoordForMeshPrimitive`
     * can be used.
     *
     * If the given value is not valid, then a validation error will
     * be added to the given context, and `false` will be returned.
     *
     * @param path - The path for validation issues
     * @param texCoord - The the texture coordinate set index, used for
     * constructing the `TEXCOORD_${texCoord}` attribute name.
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    static validateTexCoordDefinition(path: string, texCoord: number, context: ValidationContext): boolean;
    /**
     * Validate the `texCoord` definition of a texture.
     *
     * This will check whether the given value is a nonnegative integer,
     * and whether the resulting `TEXCOORD_${texCoord}` attribute
     * appears in the given mesh primitive.
     *
     * If the given value is not valid, then a validation error will
     * be added to the given context, and `false` will be returned.
     *
     * @param path - The path for validation issues
     * @param texCoord - The the texture coordinate set index, used for
     * constructing the `TEXCOORD_${texCoord}` attribute name.
     * @param gltf - The glTF object
     * @param meshPrimitive - The mesh primitive that contains the extension
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    static validateTexCoordForMeshPrimitive(path: string, texCoord: number, meshPrimitive: any, context: ValidationContext): boolean;
    /**
     * Validate the given `channels`, which may be part of a feature ID
     * texture or a property texture.
     *
     * In both cases, the `channels` must be an array with at least 1
     * element, and all elements must be nonnegative integers.
     *
     * This only performs the structural validation. The validity of
     * the elements of the `channels` array for a given number of
     * image channels can be checked with `validateChannelsForImage`.
     *
     * The validity of the binary data that is supposed to be accessed
     * using these channels is not checked here.
     *
     * If the channels are not valid, then a validation error will be
     * added to the given context, and `false` will be returned.
     *
     * @param path - The path for validation issues
     * @param channels - The channels
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    static validateChannels(path: string, channels: any, context: ValidationContext): boolean;
    /**
     * Validate the given `channels`, which may be part of a feature ID
     * texture or a property texture, for an image that has the
     * given number of channels.
     *
     * If the channels are not valid, then a validation error will be
     * added to the given context, and `false` will be returned.
     *
     * @param path - The path for validation issues
     * @param name - A name for the source of the `channels` definition.
     * This may be "feature ID texture" or "property texture property".
     * @param channels - The channels
     * @param channelsInImage - The number of channels in the image
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    static validateChannelsForImage(path: string, name: string, channels: number[], channelsInImage: number, context: ValidationContext): boolean;
}
//# sourceMappingURL=TextureValidator.d.ts.map