"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BinaryPropertyTableValuesValidator = void 0;
const _3d_tiles_tools_1 = require("3d-tiles-tools");
const _3d_tiles_tools_2 = require("3d-tiles-tools");
const RangeIterables_1 = require("./RangeIterables");
const BinaryPropertyTableEnumMetadataPropertyModel_1 = require("./BinaryPropertyTableEnumMetadataPropertyModel");
const MetadataPropertyValuesValidator_1 = require("./MetadataPropertyValuesValidator");
const BinaryPropertyTableDefaultMetadataPropertyModel_1 = require("./BinaryPropertyTableDefaultMetadataPropertyModel");
/**
 * A class for the validation of values that are stored
 * in binary property tables.
 *
 * The methods in this class assume that the structural
 * validity of the input objects has already been checked
 * by a `PropertyTableValidator` and a
 * `BinaryPropertyTableValidator`.
 *
 * @internal
 */
class BinaryPropertyTableValuesValidator {
    /**
     * Performs the validation to ensure that the given
     * `BinaryPropertyTable` contains valid values.
     *
     * @param path - The path for the `ValidationIssue` instances
     * @param binaryPropertyTable - The object to validate
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the values in the object have been valid
     */
    static validateBinaryPropertyTableValues(path, binaryPropertyTable, context) {
        let result = true;
        const binaryMetadata = binaryPropertyTable.binaryMetadata;
        const metadataClass = binaryMetadata.metadataClass;
        const classProperties = (0, _3d_tiles_tools_1.defaultValue)(metadataClass.properties, {});
        for (const propertyName of Object.keys(classProperties)) {
            const classProperty = classProperties[propertyName];
            const propertyPath = path + "/properties/" + propertyName;
            if (!BinaryPropertyTableValuesValidator.validateBinaryPropertyTablePropertyValues(propertyPath, propertyName, classProperty, binaryPropertyTable, context)) {
                result = false;
            }
        }
        return result;
    }
    /**
     * Validate the values of a single property of a `BinaryPropertyTable`
     *
     * @param path - The path of the `PropertyTablePropery`, for
     * `ValidationIssue` instances
     * @param propertyName - The property name
     * @param classProperty - The `ClassProperty`
     * @param binaryPropertyTable - The `BinaryPropertyTable`
     * @param context - The `ValidationContext`
     * @returns Whether the property is valid
     */
    static validateBinaryPropertyTablePropertyValues(path, propertyName, classProperty, binaryPropertyTable, context) {
        let result = true;
        const propertyTable = binaryPropertyTable.propertyTable;
        const propertyTableProperties = (0, _3d_tiles_tools_1.defaultValue)(propertyTable.properties, {});
        const propertyTablePropertry = propertyTableProperties[propertyName];
        const keys = RangeIterables_1.RangeIterables.range1D(propertyTable.count);
        // Perform the checks that only apply to ENUM types,
        if (classProperty.type === "ENUM") {
            const binaryMetadata = binaryPropertyTable.binaryMetadata;
            const binaryEnumInfo = binaryMetadata.binaryEnumInfo;
            const enumType = classProperty.enumType;
            const enumValueValueNames = binaryEnumInfo.enumValueValueNames[enumType];
            const enumValueNameValues = binaryEnumInfo.enumValueNameValues[enumType];
            const validEnumValueValues = Object.values(enumValueNameValues);
            const metadataPropertyModel = new BinaryPropertyTableEnumMetadataPropertyModel_1.BinaryPropertyTableEnumMetadataPropertyModel(binaryPropertyTable, propertyName, classProperty, enumValueValueNames);
            if (!MetadataPropertyValuesValidator_1.MetadataPropertyValuesValidator.validateEnumValues(path, propertyName, keys, metadataPropertyModel, validEnumValueValues, context)) {
                result = false;
            }
        }
        // Perform the checks that only apply to numeric types
        if (_3d_tiles_tools_2.ClassProperties.hasNumericType(classProperty)) {
            const metadataPropertyModel = new BinaryPropertyTableDefaultMetadataPropertyModel_1.BinaryPropertyTableDefaultMetadataPropertyModel(binaryPropertyTable, propertyName, propertyTablePropertry, classProperty);
            if (!MetadataPropertyValuesValidator_1.MetadataPropertyValuesValidator.validateMinMax(path, propertyName, keys, metadataPropertyModel, propertyTablePropertry, "property table", classProperty, context)) {
                result = false;
            }
        }
        return result;
    }
}
exports.BinaryPropertyTableValuesValidator = BinaryPropertyTableValuesValidator;
