import { ClassProperty } from "3d-tiles-tools";
import { ValidationContext } from "../ValidationContext";
import { MetadataPropertyModel } from "./MetadataPropertyModel";
/**
 * A class for the validation of values that are stored
 * as metadata properties.
 *
 * The methods in this class assume that the structural
 * validity of the input objects has already been checked.
 *
 * @internal
 */
export declare class MetadataPropertyValuesValidator {
    /**
     * Validate that the values of the specified ENUM property are valid.
     *
     * This will iterate over the given keys, obtain the RAW (numeric)
     * value for the respective key from the given model, and check
     * whether this value is contained in the given set of valid values.
     *
     * @param path - The path for `ValidationIssue` instances
     * @param propertyName - The property name
     * @param keys - The iterable of the keys that will be used for
     * accessing the metadata property model
     * @param metadataPropertyModel - The `MetadataPropertyModel`
     * @param validEnumValueValues - The values that appeared as
     * `enum.values[i].value` in the enum definition
     * @param context - The `ValidationContext`
     * @returns Whether the enum values have been valid
     */
    static validateEnumValues<T>(path: string, propertyName: string, keys: Iterable<T>, metadataPropertyModel: MetadataPropertyModel<T>, validEnumValueValues: number[], context: ValidationContext): boolean;
    /**
     * Validate the that the values of the given property model
     * are in the range that is defined by the `min`/`max`
     * values of the class property, or the property itself.
     *
     * @param path - The path for `ValidationIssue` instances
     * @param propertyName - The property name
     * definition, for example, 'class property' or 'property texture property'.
     * @param keys - The iterable of the keys that will be used for
     * accessing the metadata property model
     * @param metadataPropertyModel - The `MetadataPropertyModel`
     * @param property - The property, which may contain `min` and
     * `max` properties, and `scale` or `offset` properties that
     * override the respective value from the class property
     * @param propertySourceInfo - A string indicating the source of
     * the property value, e.g. 'property table', or 'property texture'
     * @param classProperty - The class property
     * @param context - The `ValidationContext`
     * @returns Whether the values obeyed the limit
     */
    static validateMinMax<T>(path: string, propertyName: string, keys: Iterable<T>, metadataPropertyModel: MetadataPropertyModel<T>, property: {
        max?: any;
        min?: any;
        scale?: any;
        offset?: any;
    }, propertySourceInfo: string, classProperty: ClassProperty, context: ValidationContext): boolean;
    /**
     * Validate the that none of the values of the given
     * property model is smaller than the given defined
     * minimum.
     *
     * @param path - The path for `ValidationIssue` instances
     * @param propertyName - The property name
     * @param definedMin - The defined minimum
     * @param definedMinInfo - A string indicating the source of the minimum
     * definition, for example, 'class property' or 'property texture property'.
     * @param keys - The iterable of the keys that will be used for
     * accessing the metadata property model
     * @param metadataPropertyModel - The `MetadataPropertyModel`
     * @param property - The property, which may contain `scale` or
     * `offset` properties that override the respective value from
     * the class property
     * @param propertySourceInfo - A string indicating the source of
     * the property value, e.g. 'property table', or 'property texture'
     * @param classProperty - The class property
     * @param context - The `ValidationContext`
     * @returns Whether the values obeyed the limit
     */
    private static validateMin;
    /**
     * Compute the minimum value that appears in the given
     * metadata property model.
     *
     * This assumes that the property has a numeric type,
     * as indicated by `ClassProperties.hasNumericType`.
     *
     * @param keys - The iterable of the keys that will be used for
     * accessing the metadata property model
     * @param metadataPropertyModel - The `MetadataPropertyModel`
     * @returns The minimum
     */
    private static computeMin;
    /**
     * Validate the that none of the values of the given
     * property model is greater than the given defined
     * maximum.
     *
     * @param path - The path for `ValidationIssue` instances
     * @param propertyName - The property name
     * @param definedMax - The defined maximum
     * @param definedMaxInfo - A string indicating the source of the maximum
     * definition, for example, 'class property' or 'property texture property'.
     * @param keys - The iterable of the keys that will be used for
     * accessing the metadata property model
     * @param metadataPropertyModel - The `MetadataPropertyModel`
     * @param property - The property, which may contain `scale` or
     * `offset` properties that override the respective value from
     * the class property
     * @param propertySourceInfo - A string indicating the source of
     * the property value, e.g. 'property table', or 'property texture'
     * @param classProperty - The class property
     * @param context - The `ValidationContext`
     * @returns Whether the values obeyed the limit
     */
    private static validateMax;
    /**
     * Compute the maximum value that appears in the given
     * metadata property model.
     *
     * This assumes that the property has a numeric type,
     * as indicated by `ClassProperties.hasNumericType`.
     *
     * @param keys - The iterable of the keys that will be used for
     * accessing the metadata property model
     * @param metadataPropertyModel - The `MetadataPropertyModel`
     * @returns The maximum
     */
    private static computeMax;
}
//# sourceMappingURL=MetadataPropertyValuesValidator.d.ts.map