"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MetadataPropertyValuesValidator = void 0;
const _3d_tiles_tools_1 = require("3d-tiles-tools");
const _3d_tiles_tools_2 = require("3d-tiles-tools");
const BasicValidator_1 = require("../BasicValidator");
const MetadataValueValidationMessages_1 = require("./MetadataValueValidationMessages");
const MetadataValidationIssues_1 = require("../../issues/MetadataValidationIssues");
/**
 * A class for the validation of values that are stored
 * as metadata properties.
 *
 * The methods in this class assume that the structural
 * validity of the input objects has already been checked.
 *
 * @internal
 */
class MetadataPropertyValuesValidator {
    /**
     * Validate that the values of the specified ENUM property are valid.
     *
     * This will iterate over the given keys, obtain the RAW (numeric)
     * value for the respective key from the given model, and check
     * whether this value is contained in the given set of valid values.
     *
     * @param path - The path for `ValidationIssue` instances
     * @param propertyName - The property name
     * @param keys - The iterable of the keys that will be used for
     * accessing the metadata property model
     * @param metadataPropertyModel - The `MetadataPropertyModel`
     * @param validEnumValueValues - The values that appeared as
     * `enum.values[i].value` in the enum definition
     * @param context - The `ValidationContext`
     * @returns Whether the enum values have been valid
     */
    static validateEnumValues(path, propertyName, keys, metadataPropertyModel, validEnumValueValues, context) {
        let result = true;
        for (const key of keys) {
            // The validation happens based on the RAW property
            // values for enums, i.e. the `enum.value[i].value`
            // values (which are not translated into strings)
            const rawPropertyValue = metadataPropertyModel.getRawPropertyValue(key);
            // For arrays, simply validate each element individually
            if (Array.isArray(rawPropertyValue)) {
                for (let i = 0; i < rawPropertyValue.length; i++) {
                    const rawPropertyValueElement = rawPropertyValue[i];
                    if (!BasicValidator_1.BasicValidator.validateEnum(path, propertyName + `[${key}][${i}]`, rawPropertyValueElement, validEnumValueValues, context)) {
                        result = false;
                    }
                }
            }
            else {
                if (!BasicValidator_1.BasicValidator.validateEnum(path, propertyName + `[${key}]`, rawPropertyValue, validEnumValueValues, context)) {
                    result = false;
                }
            }
        }
        return result;
    }
    /**
     * Validate the that the values of the given property model
     * are in the range that is defined by the `min`/`max`
     * values of the class property, or the property itself.
     *
     * @param path - The path for `ValidationIssue` instances
     * @param propertyName - The property name
     * definition, for example, 'class property' or 'property texture property'.
     * @param keys - The iterable of the keys that will be used for
     * accessing the metadata property model
     * @param metadataPropertyModel - The `MetadataPropertyModel`
     * @param property - The property, which may contain `min` and
     * `max` properties, and `scale` or `offset` properties that
     * override the respective value from the class property
     * @param propertySourceInfo - A string indicating the source of
     * the property value, e.g. 'property table', or 'property texture'
     * @param classProperty - The class property
     * @param context - The `ValidationContext`
     * @returns Whether the values obeyed the limit
     */
    static validateMinMax(path, propertyName, keys, metadataPropertyModel, property, propertySourceInfo, classProperty, context) {
        let result = true;
        // When the ClassProperty defines a minimum, then the metadata
        // values MUST not be smaller than this minimum
        if ((0, _3d_tiles_tools_1.defined)(classProperty.min)) {
            if (!MetadataPropertyValuesValidator.validateMin(path, propertyName, classProperty.min, "class property", keys, metadataPropertyModel, property, propertySourceInfo, classProperty, context)) {
                result = false;
            }
        }
        // When the property defines a minimum, then the metadata
        // values MUST not be smaller than this minimum
        if ((0, _3d_tiles_tools_1.defined)(property.min)) {
            const definedMin = property.min;
            if (!MetadataPropertyValuesValidator.validateMin(path, propertyName, definedMin, "property", keys, metadataPropertyModel, property, propertySourceInfo, classProperty, context)) {
                result = false;
            }
            else {
                // When none of the values is smaller than the minimum from
                // the property, make sure that this minimum matches the
                // computed minimum of all metadata values
                const computedMin = MetadataPropertyValuesValidator.computeMin(keys, metadataPropertyModel);
                if (!_3d_tiles_tools_2.ArrayValues.deepEquals(computedMin, definedMin)) {
                    const message = `For property '${propertyName}', the property ` +
                        `defines a minimum of ${definedMin}, but the computed ` +
                        `minimum value is ${computedMin}`;
                    const issue = MetadataValidationIssues_1.MetadataValidationIssues.METADATA_VALUE_MISMATCH(path, message);
                    context.addIssue(issue);
                    result = false;
                }
            }
        }
        // When the ClassProperty defines a maximum, then the metadata
        // values MUST not be greater than this maximum
        if ((0, _3d_tiles_tools_1.defined)(classProperty.max)) {
            if (!MetadataPropertyValuesValidator.validateMax(path, propertyName, classProperty.max, "class property", keys, metadataPropertyModel, property, propertySourceInfo, classProperty, context)) {
                result = false;
            }
        }
        // When the property defines a maximum, then the metadata
        // values MUST not be greater than this maximum
        if ((0, _3d_tiles_tools_1.defined)(property.max)) {
            const definedMax = property.max;
            if (!MetadataPropertyValuesValidator.validateMax(path, propertyName, definedMax, "property texture property", keys, metadataPropertyModel, property, propertySourceInfo, classProperty, context)) {
                result = false;
            }
            else {
                // When none of the values is greater than the maximum from
                // the property, make sure that this maximum matches the
                // computed maximum of all metadata values
                const computedMax = MetadataPropertyValuesValidator.computeMax(keys, metadataPropertyModel);
                if (!_3d_tiles_tools_2.ArrayValues.deepEquals(computedMax, definedMax)) {
                    const message = `For property '${propertyName}', the property ` +
                        `defines a maximum of ${definedMax}, but the computed ` +
                        `maximum value is ${computedMax}`;
                    const issue = MetadataValidationIssues_1.MetadataValidationIssues.METADATA_VALUE_MISMATCH(path, message);
                    context.addIssue(issue);
                    result = false;
                }
            }
        }
        return result;
    }
    /**
     * Validate the that none of the values of the given
     * property model is smaller than the given defined
     * minimum.
     *
     * @param path - The path for `ValidationIssue` instances
     * @param propertyName - The property name
     * @param definedMin - The defined minimum
     * @param definedMinInfo - A string indicating the source of the minimum
     * definition, for example, 'class property' or 'property texture property'.
     * @param keys - The iterable of the keys that will be used for
     * accessing the metadata property model
     * @param metadataPropertyModel - The `MetadataPropertyModel`
     * @param property - The property, which may contain `scale` or
     * `offset` properties that override the respective value from
     * the class property
     * @param propertySourceInfo - A string indicating the source of
     * the property value, e.g. 'property table', or 'property texture'
     * @param classProperty - The class property
     * @param context - The `ValidationContext`
     * @returns Whether the values obeyed the limit
     */
    static validateMin(path, propertyName, definedMin, definedMinInfo, keys, metadataPropertyModel, property, propertySourceInfo, classProperty, context) {
        let result = true;
        for (const key of keys) {
            const rawPropertyValue = metadataPropertyModel.getRawPropertyValue(key);
            const propertyValue = metadataPropertyModel.getPropertyValue(key);
            if (_3d_tiles_tools_2.ArrayValues.anyDeepLessThan(propertyValue, definedMin)) {
                const valueMessagePart = MetadataValueValidationMessages_1.MetadataValuesValidationMessages.createValueMessagePart(rawPropertyValue, classProperty, property, propertyValue);
                const message = `For property '${propertyName}', the ${definedMinInfo} ` +
                    `defines a minimum of ${definedMin}, but the value at ` +
                    `${key} of the ${propertySourceInfo} is ${valueMessagePart}`;
                const issue = MetadataValidationIssues_1.MetadataValidationIssues.METADATA_VALUE_NOT_IN_RANGE(path, message);
                context.addIssue(issue);
                result = false;
            }
        }
        return result;
    }
    /**
     * Compute the minimum value that appears in the given
     * metadata property model.
     *
     * This assumes that the property has a numeric type,
     * as indicated by `ClassProperties.hasNumericType`.
     *
     * @param keys - The iterable of the keys that will be used for
     * accessing the metadata property model
     * @param metadataPropertyModel - The `MetadataPropertyModel`
     * @returns The minimum
     */
    static computeMin(keys, metadataPropertyModel) {
        let computedMin = undefined;
        for (const key of keys) {
            const propertyValue = metadataPropertyModel.getPropertyValue(key);
            if (!(0, _3d_tiles_tools_1.defined)(computedMin)) {
                computedMin = _3d_tiles_tools_2.ArrayValues.deepClone(propertyValue);
            }
            else {
                computedMin = _3d_tiles_tools_2.ArrayValues.deepMin(computedMin, propertyValue);
            }
        }
        return computedMin;
    }
    /**
     * Validate the that none of the values of the given
     * property model is greater than the given defined
     * maximum.
     *
     * @param path - The path for `ValidationIssue` instances
     * @param propertyName - The property name
     * @param definedMax - The defined maximum
     * @param definedMaxInfo - A string indicating the source of the maximum
     * definition, for example, 'class property' or 'property texture property'.
     * @param keys - The iterable of the keys that will be used for
     * accessing the metadata property model
     * @param metadataPropertyModel - The `MetadataPropertyModel`
     * @param property - The property, which may contain `scale` or
     * `offset` properties that override the respective value from
     * the class property
     * @param propertySourceInfo - A string indicating the source of
     * the property value, e.g. 'property table', or 'property texture'
     * @param classProperty - The class property
     * @param context - The `ValidationContext`
     * @returns Whether the values obeyed the limit
     */
    static validateMax(path, propertyName, definedMax, definedMaxInfo, keys, metadataPropertyModel, property, propertySourceInfo, classProperty, context) {
        let result = true;
        for (const key of keys) {
            const rawPropertyValue = metadataPropertyModel.getRawPropertyValue(key);
            const propertyValue = metadataPropertyModel.getPropertyValue(key);
            if (_3d_tiles_tools_2.ArrayValues.anyDeepGreaterThan(propertyValue, definedMax)) {
                const valueMessagePart = MetadataValueValidationMessages_1.MetadataValuesValidationMessages.createValueMessagePart(rawPropertyValue, classProperty, property, propertyValue);
                const message = `For property '${propertyName}', the ${definedMaxInfo} ` +
                    `defines a maximum of ${definedMax}, but the value at ` +
                    `${key} of the ${propertySourceInfo} is ${valueMessagePart}`;
                const issue = MetadataValidationIssues_1.MetadataValidationIssues.METADATA_VALUE_NOT_IN_RANGE(path, message);
                context.addIssue(issue);
                result = false;
            }
        }
        return result;
    }
    /**
     * Compute the maximum value that appears in the given
     * metadata property model.
     *
     * This assumes that the property has a numeric type,
     * as indicated by `ClassProperties.hasNumericType`.
     *
     * @param keys - The iterable of the keys that will be used for
     * accessing the metadata property model
     * @param metadataPropertyModel - The `MetadataPropertyModel`
     * @returns The maximum
     */
    static computeMax(keys, metadataPropertyModel) {
        let computedMax = undefined;
        for (const key of keys) {
            const propertyValue = metadataPropertyModel.getPropertyValue(key);
            if (!(0, _3d_tiles_tools_1.defined)(computedMax)) {
                computedMax = _3d_tiles_tools_2.ArrayValues.deepClone(propertyValue);
            }
            else {
                computedMax = _3d_tiles_tools_2.ArrayValues.deepMax(computedMax, propertyValue);
            }
        }
        return computedMax;
    }
}
exports.MetadataPropertyValuesValidator = MetadataPropertyValuesValidator;
