import { ClassProperty } from "3d-tiles-tools";
import { Schema } from "3d-tiles-tools";
/**
 * Internal utilities related to the validation of ENUM values.
 *
 * These methods assume that the structural validity of their
 * inputs has already been checked. They are only utility
 * methods for the subsequent validation steps.
 *
 * @internal
 */
export declare class MetadataValidationUtilities {
    /**
     * Computes the enum 'valueType' of the specified ENUM property.
     *
     * If any required element of the input is `undefined`, then
     * `undefined` is returned.
     *
     * @param schema - The schema
     * @param className - The class name
     * @param propertyName - The property name
     * @returns The value type
     */
    static computeEnumValueType(schema: Schema, className: string, propertyName: string): string | undefined;
    /**
     * Computes the mapping from enum value values to enum value
     * names for the specified property.
     *
     * If any required element of the input is `undefined`, then
     * `undefined` is returned.
     *
     * @param schema - The schema
     * @param className - The class name
     * @param propertyName - The property name
     * @returns The value value names
     */
    static computeEnumValueValueNames(schema: Schema, className: string, propertyName: string): {
        [key: number]: string;
    } | undefined;
    /**
     * Computes the `ClassProperty` that describes the specified
     * property.
     *
     * This will find the `class` for `schema.classes[className]`,
     * and the `property` for `class[propertyName]`.
     *
     * This assumes that the given structures are valid. If any
     * element in this process is `undefined`, then `undefined`
     * is returned.
     *
     * @param schema - The metadata schema
     * @param className - The class name
     * @param propertyName - The property name
     * @returns The valid enum value values
     */
    static computeClassProperty(schema: Schema, className: string, propertyName: string): ClassProperty | undefined;
    /**
     * Computes the set of (numeric) enum values that are valid
     * for the specified property.
     *
     * This will find...
     * - the `class` for `schema.classes[className]`
     * - the `property` for `class[propertyName]`
     * - the `enum` for `schema.enums[property.enumType]`
     * - all values that appear as `enum.values[i].value`
     *
     * This assumes that the given structures are valid. If any
     * element in this process is `undefined`, then `undefined`
     * is returned.
     *
     * @param schema - The metadata schema
     * @param className - The class name
     * @param propertyName - The property name
     * @returns The valid enum value values
     */
    static computeValidEnumValueValues(schema: Schema, className: string, propertyName: string): number[] | undefined;
    /**
     * Returns the metadata enum with the given name from the
     * given schema, or `undefined` if any required element
     * is `undefined`
     *
     * @param schema - The schema
     * @param enumType - The enum type
     * @returns The metadata enum
     */
    private static computeMetadataEnum;
}
//# sourceMappingURL=MetadataValidationUtilities.d.ts.map