"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MetadataValuesValidationMessages = void 0;
const _3d_tiles_tools_1 = require("3d-tiles-tools");
/**
 * Utility methods related to the messages that are part of
 * validation issues in the metadata validation.
 */
class MetadataValuesValidationMessages {
    /**
     * Creates a message that describes how a metadata value was computed.
     *
     * The intention is to insert this as `The value is ${valueMessagePart}`
     * in a message that explains how the `value` was computed from the
     * raw value, normalization, offset and scale.
     *
     * @param rawValue - The raw value (e.g. from the `PropertyModel`),
     * without normalization, offset, or scale
     * @param classProperty - The class property
     * @param property - The property, which may contain `scale` or
     * `offset` properties that override the respective value from
     * the class property
     * @param value - The final value
     * @returns The message part
     */
    static createValueMessagePart(rawValue, classProperty, property, value) {
        // Determine the scale and its source (i.e whether it
        // is defined in the class property, or overridden
        // in the actual property)
        let scale = undefined;
        let scaleSource = undefined;
        if ((0, _3d_tiles_tools_1.defined)(property.scale)) {
            scale = property.scale;
            scaleSource = "property.scale";
        }
        else if ((0, _3d_tiles_tools_1.defined)(classProperty.scale)) {
            scale = classProperty.scale;
            scaleSource = "classProperty.scale";
        }
        // Determine the offset and its source (i.e whether it
        // is defined in the class property, or overridden
        // in the actual property)
        let offset = undefined;
        let offsetSource = undefined;
        if ((0, _3d_tiles_tools_1.defined)(property.offset)) {
            offset = property.offset;
            offsetSource = "property.offset";
        }
        else if ((0, _3d_tiles_tools_1.defined)(classProperty.offset)) {
            offset = classProperty.offset;
            offsetSource = "classProperty.offset";
        }
        const normalized = classProperty.normalized;
        const componentType = classProperty.componentType;
        if ((0, _3d_tiles_tools_1.defined)(offset) && (0, _3d_tiles_tools_1.defined)(scale)) {
            if (normalized === true) {
                const messagePart = `computed as normalize${componentType}(rawValue)*${scaleSource}+${offsetSource} ` +
                    `= normalize${componentType}(${rawValue})*${scale}+${offset} = ${value}`;
                return messagePart;
            }
            const messagePart = `computed as rawValue*${scaleSource}+${offsetSource} ` +
                `= ${rawValue}*${scale}+${offset} = ${value}`;
            return messagePart;
        }
        if ((0, _3d_tiles_tools_1.defined)(offset)) {
            if (normalized === true) {
                const messagePart = `computed as normalize${componentType}(rawValue)+${offsetSource} ` +
                    `= normalize${componentType}(${rawValue})+${offset} = ${value}`;
                return messagePart;
            }
            const messagePart = `computed as rawValue+${offsetSource} ` +
                `= ${rawValue}+${offset} = ${value}`;
            return messagePart;
        }
        if ((0, _3d_tiles_tools_1.defined)(scale)) {
            if (normalized === true) {
                const messagePart = `computed as normalize${componentType}(rawValue)*${scaleSource} ` +
                    `= normalize${componentType}(${rawValue})*${scale} = ${value}`;
                return messagePart;
            }
            const messagePart = `computed as rawValue*${scaleSource} ` +
                `= ${rawValue}*${scale} = ${value}`;
            return messagePart;
        }
        if (normalized === true) {
            const messagePart = `computed as normalize${componentType}(rawValue) ` +
                `= normalize${componentType}(${rawValue}) = ${value}`;
            return messagePart;
        }
        const messagePart = `${value}`;
        return messagePart;
    }
}
exports.MetadataValuesValidationMessages = MetadataValuesValidationMessages;
