import { Schema } from "3d-tiles-tools";
import { ValidationContext } from "../ValidationContext";
import { ValidatedElement } from "../ValidatedElement";
/**
 * Utility class for validating the definition of a metadata schema.
 *
 * The metadata schema can either be defined as an inlined `schema`
 * object in the JSON, or via a `schemUri` (but not both!).
 * This class resolves and validates the schema from either of
 * these sources, and returns information about that validation
 * result.
 */
export declare class SchemaDefinitionValidator {
    /**
     * Validates the given schema definition.
     *
     * The returned object will contain two properties:
     * - `wasPresent`: Whether any schema definition was given
     * - `validatedElement`: The validated `Schema` object
     *
     * If neither `schema` nor `schemaUri` are given, then the result
     * will be `{false, undefined}`.
     *
     * If the `schema` and `schemaUri` are both given, then an error will be
     * added to the given context, and `{ true, undefined }` will be returned.
     *
     * If the `schemaUri` is given but invalid, then an error will be
     * added to the given context, and `{ true, undefined }` will be returned.
     *
     * If the `schema` was given, or the schema could be resolved from
     * the `schemaUri`, and the schema turned out to be valid, then
     * `{ true, validatedSchema }` will be returned.
     *
     * @param path - The path for `ValidationIssue` instances
     * @param name - A name for the containing object (usually 'tileset')
     * @param schema - The `schema` object from the JSON
     * @param schemaUri - The `schemaUri` from the JSON
     * @param context - The `ValidatonContext`
     * @returns The schema definition validation result
     */
    static validateSchemaDefinition(path: string, name: string, schema: any, schemaUri: any, context: ValidationContext): Promise<ValidatedElement<Schema>>;
    /**
     * Resolves the schema from the given URI
     *
     * @param path - The path for validation issues
     * @param schemaUri - The schema URI
     * @param context - The `ValidationContext`
     * @returns A promise that resolves with the result object,
     * or `undefined` if the schema could not be resolved
     */
    private static resolveSchema;
}
//# sourceMappingURL=SchemaDefinitionValidator.d.ts.map